import logging
import os

logger = logging.getLogger("qondor")


def cmsconnect_get_all_sites():
    """
    Reads the central config for cmsconnect to determine the list of all available sites
    """
    try:
        from configparser import RawConfigParser  # python 3
    except ImportError:
        import ConfigParser  # python 2

        RawConfigParser = ConfigParser.RawConfigParser
    cfg = RawConfigParser()
    cfg.read("/etc/ciconnect/config.ini")
    sites = cfg.get("submit", "DefaultSites").split(",")
    all_sites = set(sites)
    return all_sites


def cmsconnect_settings(sub, blacklist=None, whitelist=None, cli=False):
    """
    Adds special cmsconnect settings to submission dict in order to submit
    on cmsconnect via the python bindings, or in case of submitting by the
    command line, to set the DESIRED_Sites key.
    Modifies the dict in place.
    """
    all_sites = cmsconnect_get_all_sites()

    # Check whether the user whitelisted or blacklisted some sites
    desired_sites = None
    if blacklist or whitelist:
        import fnmatch

        blacklisted = []
        whitelisted = []
        # Build the blacklist
        if blacklist:
            for blacksite_pattern in blacklist:
                for site in all_sites:
                    if fnmatch.fnmatch(site, blacksite_pattern):
                        blacklisted.append(site)
        # Build the whitelist
        if whitelist:
            for whitesite_pattern in whitelist:
                for site in all_sites:
                    if fnmatch.fnmatch(site, whitesite_pattern):
                        whitelisted.append(site)
        # Convert to list and sort
        blacklisted = list(set(blacklisted))
        blacklisted.sort()
        whitelisted = list(set(whitelisted))
        whitelisted.sort()
        logger.info("Blacklisting: %s", ",".join(blacklisted))
        logger.info("Whitelisting: %s", ",".join(whitelisted))
        desired_sites = list(
            (set(all_sites) - set(blacklisted)).union(set(whitelisted))
        )
        desired_sites.sort()

    # Add a plus only if submitting via .jdl file
    def addplus(key):
        return "+" + key if cli else key

    if desired_sites:
        logger.info("Submitting to desired sites: %s", ",".join(desired_sites))
        sub[addplus("DESIRED_Sites")] = '"' + ",".join(desired_sites) + '"'
    else:
        logger.info("Submitting to all sites: %s", ",".join(all_sites))
    if not cli:
        sub[addplus("ConnectWrapper")] = '"2.0"'
        sub[addplus("CMSGroups")] = '"/cms,T3_US_FNALLPC"'
        sub[addplus("MaxWallTimeMins")] = "500"
        sub[addplus("ProjectName")] = '"cms.org.fnal"'
        sub[addplus("SubmitFile")] = '"irrelevant.jdl"'
        sub[addplus("AccountingGroup")] = '"analysis.{0}"'.format(os.environ["USER"])
        logger.warning(
            "FIXME: CMS Connect settings currently hard-coded for a FNAL user"
        )



import logging

COLORS = {
    "yellow": "\033[33m",
    "red": "\033[31m",
    "green": "\033[32m",
}
RESET = "\033[0m"


def colored(text, color=None):
    if color is not None:
        text = COLORS[color] + text + RESET
    return text


def setup_logger(name="qondor", fmt=None):
    if name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(name)
        logger.info("Logger %s is already defined", name)
    else:
        if fmt is None:
            fmt = logging.Formatter(
                fmt=(
                    colored(
                        "[{0}|%(levelname)8s|%(asctime)s|%(module)s]:".format(name),
                        "yellow",
                    )
                    + " %(message)s"
                ),
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
    return logger


def setup_subprocess_logger():
    return setup_logger(
        "subprocess",
        fmt=logging.Formatter(
            fmt=colored("[%(asctime)s]:", "red") + " %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        ),
    )



from setuptools import setup

with open("qondor/include/VERSION", "r") as f:
    version = f.read().strip()

setup(
    name="qondor",
    version=version,
    license="BSD 3-Clause License",
    description="Description text",
    url="https://github.com/tklijnsma/qondor.git",
    author="Thomas Klijnsma",
    author_email="tklijnsm@gmail.com",
    packages=["qondor"],
    zip_safe=False,
    scripts=[
        "bin/qondor-submit",
        "bin/qondor-resubmit",
        "bin/qondor-status",
        "bin/qondor-make-cmssw-tarball",
        "bin/qondor-version",
    ],
    install_requires=["seutils"],
    include_package_data=True,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import pprint
import re
from contextlib import contextmanager
from time import sleep, strftime

import qondor

logger = logging.getLogger("qondor")


CACHED_FUNCTIONS = []


def cache_return_value(func):
    """
    Decorator that only calls a function once, and
    subsequent calls just return the cached return value
    """
    global CACHED_FUNCTIONS

    def wrapper(*args, **kwargs):
        if not getattr(wrapper, "is_called", False):
            wrapper.is_called = True
            wrapper.cached_return_value = func(*args, **kwargs)
            CACHED_FUNCTIONS.append(wrapper)
        else:
            logger.debug(
                "Returning cached value for %s: %s",
                func.__name__,
                wrapper.cached_return_value,
            )
        return wrapper.cached_return_value

    return wrapper


def clear_cache():
    global CACHED_FUNCTIONS
    for func in CACHED_FUNCTIONS:
        func.is_called = False
    CACHED_FUNCTIONS = []


class ScheddManager(object):
    """
    Class that interacts with htcondor scheduler daemons ('schedds').
    Will first run `get_collector_node_addresses` to obtain htcondor collectors,
    and will get the schedd_ads from the collectors.
    The algorithm for `get_best_schedd` is based on the weight calculation
    implemented on the Fermilab HTCondor cluster.

    Most methods are decorated with `cache_return_value`, meaning subsequent calls
    will return the same value as the first call until the cache is cleared.
    """

    def __init__(self):
        super(ScheddManager, self).__init__()
        self.schedd_ads = []
        self.schedd_constraints = None

    def clear_cache(self):
        """Convenience function to not have to import clear_cache everywhere"""
        clear_cache()

    @cache_return_value
    def get_collector_node_addresses(self):
        """
        Sets the attribute self.collector_node_addresses to a list of addresses to
        htcondor collectors. The collectors subsequently are able to return
        schedulers to which one can submit.

        First looks if qondor.COLLECTOR_NODES is set to anything. If so, it takes the
        addresses from that, allowing the user to easily configure custom collector nodes.

        Otherwise, it tries to obtain addresses from the htcondor parameter COLLECTOR_HOST_STRING.
        """
        import htcondor

        if not (qondor.COLLECTOR_NODES is None):
            logger.info(
                "Setting collectors to qondor.COLLECTOR_NODES %s",
                qondor.COLLECTOR_NODES,
            )
            self.collector_node_addresses = qondor.COLLECTOR_NODES
            return

        err_msg = (
            "Could not find any collector nodes. "
            "Either set qondor.COLLECTOR_NODES manually to a ', '-separated "
            "list of addresses, or set the htcondor parameter 'COLLECTOR_HOST_STRING',"
            " or subclass qondor.schedd.ScheddManager so that it can find the collector nodes."
        )
        try:
            collector_node_addresses = (
                htcondor.param["COLLECTOR_HOST_STRING"].strip().strip('"')
            )
        except Exception:
            logger.error(err_msg)
            raise
        if collector_node_addresses is None:
            RuntimeError(err_msg)

        logger.debug("Found collector nodes %s", collector_node_addresses)
        # Seems like a very obfuscated way of writing ".split()" but keep it for now
        self.collector_node_addresses = re.findall(
            r"[\w\/\:\/\-\/\.]+", collector_node_addresses
        )
        logger.debug("Using collector_node_addresses %s", self.collector_node_addresses)

    @cache_return_value
    def get_schedd_ads(self):
        import htcondor

        schedd_ad_projection = [
            "Name",
            "MyAddress",
            "MaxJobsRunning",
            "ShadowsRunning",
            "RecentDaemonCoreDutyCycle",
            "TotalIdleJobs",
        ]
        try:
            logger.debug("First trying default collector and schedd")
            # This is most likely to work for most batch systems
            collector = htcondor.Collector()
            limited_schedd_ad = collector.locate(htcondor.DaemonTypes.Schedd)
            logger.debug(
                "Retrieved limited schedd ad:\n%s", pprint.pformat(limited_schedd_ad)
            )
            self.schedd_ads = collector.query(
                htcondor.AdTypes.Schedd,
                projection=schedd_ad_projection,
                constraint='MyAddress=?="{0}"'.format(limited_schedd_ad["MyAddress"]),
            )
        except Exception as e:
            logger.debug(
                "Default collector and schedd did not work:\n%s\nTrying via collector host string",
                e,
            )
            self.get_collector_node_addresses()
            for node in self.collector_node_addresses:
                logger.debug("Querying %s for htcondor.AdTypes.Schedd", node)
                collector = htcondor.Collector(node)
                try:
                    self.schedd_ads = collector.query(
                        htcondor.AdTypes.Schedd,
                        projection=schedd_ad_projection,
                        constraint=self.schedd_constraints,
                    )
                    if self.schedd_ads:
                        # As soon as schedd_ads are found in one collector node, use those
                        # This may not be the correct choice for some batch systems
                        break
                except Exception as e:
                    logger.debug("Failed querying %s: %s", node, e)
                    continue
            else:
                logger.error(
                    "Failed to collect any schedds from %s",
                    self.collector_node_addresses,
                )
                raise RuntimeError

        logger.debug(
            "Found schedd ads: \n%s", pprint.pformat([dict(d) for d in self.schedd_ads])
        )
        return self.schedd_ads

    @cache_return_value
    def get_best_schedd(self):
        import htcondor

        self.get_schedd_ads()
        if len(self.schedd_ads) == 1:
            best_schedd_ad = self.schedd_ads[0]
        else:
            self.schedd_ads.sort(key=self.get_schedd_weight)
            best_schedd_ad = self.schedd_ads[0]
        logger.debug("Best schedd is %s", best_schedd_ad["Name"])
        schedd = htcondor.Schedd(best_schedd_ad)
        return schedd

    @cache_return_value
    def get_all_schedds(self):
        import htcondor

        return [htcondor.Schedd(schedd_ad) for schedd_ad in self.get_schedd_ads()]

    def get_schedd_weight(self, schedd_ad):
        duty_cycle = schedd_ad["RecentDaemonCoreDutyCycle"] * 100
        occupancy = (schedd_ad["ShadowsRunning"] / schedd_ad["MaxJobsRunning"]) * 100
        n_idle_jobs = schedd_ad["TotalIdleJobs"]
        weight = 0.7 * duty_cycle + 0.2 * occupancy + 0.1 * n_idle_jobs
        logger.debug(
            "Weight calc for %s: weight = %s, "
            "duty_cycle = %s, occupancy = %s, n_idle_jobs = %s",
            schedd_ad["Name"],
            weight,
            duty_cycle,
            occupancy,
            n_idle_jobs,
        )
        return weight


class ScheddManagerFermiHTC(ScheddManager):
    """
    Subclass of ScheddManager specifically for the Fermilab HTCondor setup
    """

    def __init__(self):
        super(ScheddManagerFermiHTC, self).__init__()
        self.schedd_constraints = (
            'FERMIHTC_DRAIN_LPCSCHEDD=?=FALSE && FERMIHTC_SCHEDD_TYPE=?="CMSLPC"'
        )

    @cache_return_value
    def get_collector_node_addresses(self):
        import htcondor

        try:
            collector_node_addresses = htcondor.param["FERMIHTC_REMOTE_POOL"]
            self.collector_node_addresses = re.findall(
                r"[\w\/\:\/\-\/\.]+", collector_node_addresses
            )
            logger.info("Set collector_node_addresses to %s", collector_node_addresses)
        except KeyError:
            super(ScheddManagerFermiHTC, self).get_collector_node_addresses()


# -------------- Convenience functions
# Define a GLOBAL_SCHEDDMAN, so that for basic usage one
# can simply call e.g. 'get_best_schedd'

GLOBAL_SCHEDDMAN_CLS = ScheddManager
GLOBAL_SCHEDDMAN = None


def _get_scheddman(renew):
    global GLOBAL_SCHEDDMAN
    if GLOBAL_SCHEDDMAN is None:
        # Create a new instance
        GLOBAL_SCHEDDMAN = GLOBAL_SCHEDDMAN_CLS()
    if renew:
        GLOBAL_SCHEDDMAN.clear_cache()
    return GLOBAL_SCHEDDMAN


def get_best_schedd(renew=False):
    return _get_scheddman(renew).get_best_schedd()


def get_schedd_ads(renew=False):
    return _get_scheddman(renew).get_schedd_ads()


def get_schedds(renew=False):
    return _get_scheddman(renew).get_all_schedds()


# _____________________________________________________________________
# Some basic condor utilities

_status_str_to_int = {
    "U": 0,  # Unexpanded
    "I": 1,  # Idle
    "R": 2,  # Running
    "X": 3,  # Removed
    "C": 4,  # Completed
    "H": 5,  # Held
    "E": 6,  # Submission_err
}

_status_int_to_str = {
    0: "Unexpanded",
    1: "Idle",
    2: "Running",
    3: "Removed",
    4: "Completed",
    5: "Held",
    6: "Submission_err",
}


class QueuedJob(object):
    """
    Simple container for a job that is currently in the queue of htcondor.
    Can be initialized both by the python-bindings or the condor_q command line
    """

    @classmethod
    def from_ad(cls, ad):
        return cls(
            ad["ClusterId"],
            ad["ProcId"],
            ad["JobStatus"],
            user=ad["Owner"],
            rundir=ad["Iwd"],
        )

    def __init__(
        self, cluster_id, proc_id, status, user=None, sh_base=None, rundir=None
    ):
        self.cluster_id = cluster_id
        self.proc_id = proc_id
        self.status = status
        self.user = user
        self.sh_base = sh_base
        self.rundir = rundir

    def __repr__(self):
        return (
            super(QueuedJob, self)
            .__repr__()
            .replace(
                "object",
                "object {}.{} {}".format(
                    self.cluster_id, self.proc_id, self.status_str()
                ),
            )
        )

    def status_str(self):
        return _status_int_to_str[self.status]


def sort_jobs(list_of_jobs):
    list_of_jobs.sort(key=lambda job: (job.cluster_id, job.proc_id))


def get_jobs_bindings(cluster_id=None, proc_id=None, user="auto"):
    requirements = []
    if cluster_id:
        requirements.append("ClusterId=={0}".format(cluster_id))
    if proc_id:
        requirements.append("ProcId == {0}".format(proc_id))
    if user == "auto":
        import getpass

        user = getpass.getuser()
    if user:
        requirements.append('Owner == "{0}"'.format(user))
    requirements = " && ".join(requirements)
    classads = []
    logger.debug("requirements = %s", requirements)
    for schedd in get_schedds():
        classads.extend(
            list(
                schedd.xquery(
                    requirements=requirements,
                    projection=["ClusterId", "ProcId", "JobStatus", "Iwd", "Owner"],
                )
            )
        )
    logger.info("Query returned %s", classads)
    jobs = [QueuedJob.from_ad(ad) for ad in classads]
    sort_jobs(jobs)
    return jobs


def remove_jobs(cluster_id):
    import htcondor

    logger.info("Removing cluster_id %s", cluster_id)
    for schedd in get_schedds():
        schedd.act(htcondor.JobAction.Remove, 'ClusterId=?="{0}"'.format(cluster_id))


def wait(cluster_id, proc_id=None, n_sleep=10):
    while True:
        states = [j["JobStatus"] for j in get_jobs(cluster_id, proc_id)]
        is_done = all([not (state in [1, 2]) for state in states])
        if is_done:
            logger.info("ClusterId %s ProcId %s seems done", cluster_id, proc_id)
            break
        else:
            logger.info(
                "ClusterId %s ProcId %s not yet done:\n%s", cluster_id, proc_id, states
            )
            logger.info("Sleeping for %s seconds before checking again", n_sleep)
            sleep(n_sleep)


def get_jobs_cli(cluster_id=None, proc_id=None):
    output = qondor.utils.run_command("condor_q", env="clean", shell=True)
    jobs = []
    for line in output:
        line = line.strip()
        if not (len(line)):
            continue
        components = line.split()
        if not re.match(r"\d+\.\d+", components[0]):
            continue
        this_cluster_id, this_proc_id = map(int, components[0].split("."))
        if not (cluster_id is None) and this_cluster_id != cluster_id:
            continue
        if not (proc_id is None) and this_proc_id != proc_id:
            continue
        status_str = components[5]
        status = _status_str_to_int.get(status_str, -1)
        sh_base = components[-1]
        user = components[1]
        jobs.append(
            QueuedJob(this_cluster_id, this_proc_id, status, user=user, sh_base=sh_base)
        )
    sort_jobs(jobs)
    return jobs


def get_jobs(*args, **kwargs):
    method = kwargs.pop("method", None)
    if method is None:
        try:
            import htcondor  # noqa F401

            return get_jobs_bindings(*args, **kwargs)
        except ImportError:
            return get_jobs_cli(*args, **kwargs)


# _____________________________________________________________________
# Submission utils


def get_default_sub():
    """
    Returns the default submission dict (the equivalent of a .jdl file)
    to be used by the submitter.
    """
    sub = {
        "universe": "vanilla",
        "output": "out_$(Cluster)_$(Process).txt",
        "error": "err_$(Cluster)_$(Process).txt",
        "log": "log_$(Cluster)_$(Process).txt",
        "should_transfer_files": "YES",
        "environment": {
            "QONDOR_BATCHMODE": "1",
            "CONDOR_CLUSTER_NUMBER": "$(Cluster)",
            "CONDOR_PROCESS_ID": "$(Process)",
            "CLUSTER_SUBMISSION_TIMESTAMP": strftime(qondor.TIMESTAMP_FMT),
        },
    }
    # Try to set some more things
    try:
        sub["x509userproxy"] = os.environ["X509_USER_PROXY"]
    except KeyError:
        try:
            sub["x509userproxy"] = qondor.utils.run_command(
                ["voms-proxy-info", "-path"]
            )[0].strip()
            logger.info(
                'Set x509userproxy to "%s" based on output from voms-proxy-info',
                sub["x509userproxy"],
            )
        except Exception:
            logger.warning(
                "Could not find a x509userproxy to pass; manually "
                "set the htcondor variable 'x509userproxy' if your "
                "htcondor setup requires it."
            )
    try:
        sub["environment"]["USER"] = os.environ["USER"]
    except KeyError:
        # No user specified, no big deal
        pass
    return sub


def _change_submitobject_env_variable(submitobject, key, value):
    """
    Tries to replace an environment variable in a Submit-object.
    This hack is needed to have items be in the same cluster.
    """
    env = submitobject["environment"]
    new_env = re.sub(key + r"=\'.*?\'", "{0}='{1}'".format(key, value), env)
    logger.debug("Replacing:\n  %s\n  by\n  %s", env, new_env)
    submitobject["environment"] = new_env


def format_env_htcondor(env):
    """
    Takes a dict of key : value pairs that are both strings, and
    returns a string that is formatted so that htcondor can turn it
    into environment variables
    """
    return (
        '"'
        + " ".join(["{0}='{1}'".format(key, value) for key, value in env.items()])
        + '"'
    )


@contextmanager
def _transaction(schedd, dry=None):
    """
    Wrapper for schedd.transaction, + the ability for dry mode
    """
    if dry is None:
        dry = qondor.DRYMODE
    try:
        if not dry:
            with schedd.transaction() as transaction:
                yield transaction
        else:
            yield None
    finally:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import logging
import math
import os
import os.path as osp
import pprint
import shutil
import subprocess
import sys
import time
from contextlib import contextmanager

import qondor

logger = logging.getLogger("qondor")
subprocess_logger = logging.getLogger("subprocess")

# Python 2.6 compatibiltiy (see https://stackoverflow.com/a/13160748/9209944)
if "check_output" not in dir(subprocess):  # duck punch it in!
    logger.warning("Duck punching subprocess.check_output; suboptimal!")

    def f(*popenargs, **kwargs):
        if "stdout" in kwargs:
            raise ValueError("stdout argument not allowed, it will be overridden.")
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd)
        return output

    subprocess.check_output = f


def _create_directory_no_checks(dirname, dry=None):
    """
    Creates a directory without doing any further checks.

    :param dirname: Name of the directory to be created
    :type dirname: str
    :param dry: Don't actually create the directory, only log
    :type dry: bool, optional
    """
    if dry is None:
        dry = qondor.DRYMODE
    if not dry:
        logger.warning("Creating directory %s", dirname)
        os.makedirs(dirname)


def create_directory(dirname, renew=False, must_not_exist=False, dry=None):
    """
    Creates a directory if certain conditions are met.

    :param dirname: Name of the directory to be created
    :type dirname: str
    :param renew: Removes the directory `dirname` if it already exists
    :type renew: bool, optional
    :param must_not_exist: Throw an OSError if the directory already exists
    :type must_not_exist: bool, optional
    :param dry: Don't actually create the directory, only log
    :type dry: bool, optional
    """
    if dry is None:
        dry = qondor.DRYMODE
    if osp.isfile(dirname):
        raise OSError("{0} is a file".format(dirname))
    isdir = osp.isdir(dirname)

    if isdir:
        if must_not_exist:
            raise OSError("{0} must not exist but exists".format(dirname))
        elif renew:
            logger.warning("Deleting directory %s", dirname)
            if not dry:
                shutil.rmtree(dirname)
        else:
            logger.debug("%s already exists, not recreating", dirname)
            return
    _create_directory_no_checks(dirname, dry=dry)


def copy_file(src, dst, dry=None):
    if dry is None:
        dry = qondor.DRYMODE
    logger.info("Copying %s --> %s", src, dst)
    if not dry:
        shutil.copy(src, dst)


class switchdir(object):
    """
    Context manager to temporarily change the working directory.

    :param newdir: Directory to change into
    :type newdir: str
    :param dry: Don't actually change directory if set to True
    :type dry: bool, optional
    """

    def __init__(self, newdir, dry=None):
        super(switchdir, self).__init__()
        self.newdir = newdir
        self._backdir = os.getcwd()
        self._no_need_to_change = self.newdir == self._backdir
        self.dry = qondor.DRYMODE if dry is None else dry

    def __enter__(self):
        if self._no_need_to_change:
            logger.info("Already in right directory, no need to change")
            return
        logger.info("chdir to {0}".format(self.newdir))
        if not self.dry:
            os.chdir(self.newdir)

    def __exit__(self, type, value, traceback):
        if self._no_need_to_change:
            return
        logger.info("chdir back to {0}".format(self._backdir))
        if not self.dry:
            os.chdir(self._backdir)


def run_command(cmd, env=None, dry=None, shell=False):
    logger.warning(
        "Issuing command: {0}".format(" ".join(cmd) if not is_string(cmd) else cmd)
    )
    if dry is None:
        dry = qondor.DRYMODE
    if dry:
        return
    if shell and not is_string(cmd):
        cmd = " ".join(cmd)
    if env == "clean":
        env = get_clean_env()
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
        universal_newlines=True,
        shell=shell,
    )

    output = []
    for stdout_line in iter(process.stdout.readline, ""):
        subprocess_logger.info(stdout_line.rstrip("\n"))
        output.append(stdout_line)
    process.stdout.close()
    process.wait()
    returncode = process.returncode

    if returncode == 0:
        logger.info("Command exited with status 0 - all good")
    else:
        logger.error("Exit status {0} for command: {1}".format(returncode, cmd))
        raise subprocess.CalledProcessError(cmd, returncode)
    return output


def run_multiple_commands(cmds, env=None, dry=None):
    logger.info("Sending cmds:\n{0}".format(pprint.pformat(cmds)))
    if dry is None:
        dry = qondor.DRYMODE
    if dry:
        logger.info("Dry mode - not running command")
        return

    try:
        # Python 3 accepts encoding
        process = subprocess.Popen(
            "bash",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            bufsize=1,
            close_fds=True,
            encoding="utf8",
        )
    except TypeError:
        # Python 2 doesn't
        process = subprocess.Popen(
            "bash",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            bufsize=1,
            close_fds=True,
            universal_newlines=True,
        )

    # Break on first error (stdin will still be written but execution will be stopped)
    process.stdin.write("set -e\n")
    process.stdin.flush()

    for cmd in cmds:
        if not (is_string(cmd)):
            cmd = " ".join(cmd)
        if not (cmd.endswith("\n")):
            cmd += "\n"
        process.stdin.write(cmd)
        process.stdin.flush()
    process.stdin.close()

    output = []
    process.stdout.flush()
    for line in iter(process.stdout.readline, ""):
        if len(line) == 0:
            break
        line = line.rstrip("\n")
        subprocess_logger.info(line)
        output.append(line)

    process.stdout.close()
    process.wait()
    returncode = process.returncode

    if returncode == 0:
        logger.info("Command exited with status 0 - all good")
        return output
    else:
        raise subprocess.CalledProcessError(cmd, returncode)


def download_url_to_str(url):
    """
    Downloads a url and puts the contents in a string
    """
    logger.info("Retrieving url %s", url)
    try:
        import urllib.request

        with urllib.request.urlopen(url) as f:
            html = f.read().decode("utf-8")
    except ImportError:
        import urllib2

        response = urllib2.urlopen(url)
        html = response.read()
    return html


def iter_strip_comments(python_code):
    """
    Strips comments from python code as a string.
    Does *not* handle '#' appearing in a string
    """
    for line in python_code.split("\n"):
        line = line.split("#")[0].strip()
        if not len(line):
            continue
        yield line


def strip_comments(python_code):
    """
    As iter_strip_comments but returns a list
    """
    return "\n".join(iter_strip_comments(python_code))


class DummyFile(object):
    def write(self, text):
        pass


@contextmanager
def openfile(*args, **kwargs):
    """
    Wrapper around the standard open(...) context, with the option of drymode
    """
    dry = kwargs.pop("dry", qondor.DRYMODE)
    try:
        if dry:
            yield DummyFile()
        else:
            with open(*args, **kwargs) as f:
                yield f
    finally:
        pass


def get_exitcode(cmd):
    if is_string(cmd):
        cmd = [cmd]
    logger.debug('Getting exit code for "%s"', " ".join(cmd))
    FNULL = open(os.devnull, "w")
    if qondor.DRYMODE:
        returncode = 0
    else:
        process = subprocess.Popen(cmd, stdout=FNULL, stderr=subprocess.STDOUT)
        process.communicate()[0]
        returncode = process.returncode
    logger.debug("Got exit code %s", returncode)
    return returncode


def is_string(string):
    """
    Checks strictly whether `string` is a string
    Python 2 / 3 compatibility (https://stackoverflow.com/a/22679982/9209944)
    """
    try:
        basestring
    except NameError:
        basestring = str
    return isinstance(string, basestring)


def pip_has_version(package):
    """
    For pip install strings: Checks if there is a part that mentions the version
    """
    for c in ["<", "=", ">"]:
        if c in package:
            return True
    return False


def pip_split_version(package):
    """
    For pip install strings: Splits the part of the package name and the version part
    """
    split_at_index = False
    for c in ["<", "=", ">"]:
        if c in package:
            index = package.index(c)
            if split_at_index is False or index < split_at_index:
                split_at_index = index
    if split_at_index:
        return package[:split_at_index], package[split_at_index:]
    return package, ""


def get_installation_path_of_module(module):
    logger.debug("Trying to determine installation path of %s", module.__name__)
    try:
        logger.debug("Using pkg_resources")
        import pkg_resources

        distribution = pkg_resources.get_distribution(module.__name__)
        path = osp.abspath(distribution.location)
    except Exception:
        logger.debug("From module.__path__")
        path = osp.abspath(module.__path__[0])
        if not osp.exists(path):
            logger.warning(
                "Path %s for module %s does not exist; reloading and trying again",
                path,
                module,
            )
            try:
                from importlib import reload
            except ImportError:
                # Probably this is python 2 then
                pass
            reload(module)
            path = osp.abspath(module.__path__[0])
            if osp.exists(path):
                logger.warning(
                    "Path %s for module %s exists. "
                    "Did you chdir before calling this function? "
                    "module.__path__ is a relative path set at import time.",
                    path,
                    module,
                )
            else:
                raise RuntimeError(
                    "Found no way of determining where %s is installed", module.__name__
                )
    logger.info("Determined installation path of %s: %s", module.__name__, path)
    return path


def tarball_python_module(
    module, outdir=None, allow_uncommitted=True, dry=None, assume_pypi=True
):
    """
    Takes a python module or the name of a module, and attempts to make an installable
    pypi-style package tarball out of it.
    If assume_pypi is True, it will look for the pip installation directory of the package, and check
    whether there is a setup.py.
    Otherwise, it will look for the top-level git repository and make a tarball from that, including
    only files that are tracked by git. Uncommitted changes are included, unless allowed_uncommitted
    is set to False.
    """
    import importlib

    if dry is None:
        dry = qondor.DRYMODE
    outdir = os.getcwd() if outdir is None else outdir
    outdir = osp.abspath(outdir)

    # Input variable may be a module name
    if is_string(module):
        try:
            module = importlib.import_module(module)
        except ImportError:
            logger.error(
                "Cannot make tarball for package {}; package is not importable".format(
                    module
                )
            )
            raise
    path = get_installation_path_of_module(module)

    # Make sure path exists and is a directory
    if not osp.exists(path):
        logger.error("Path %s does not seem to exist; cwd: %s", path, os.getcwd())
        raise OSError("{0} is not a valid path".format(path))
    elif osp.isfile(path):
        path = osp.dirname(path)

    if assume_pypi:
        # Make sure it has a setup.py
        setuppy = osp.join(path, "setup.py")
        if not osp.isfile(setuppy):
            raise OSError(
                "Error installing module {0}; Expected file {1} to exist; not a valid pypi package".format(
                    module, setuppy
                )
            )
        outfile = osp.join(outdir, osp.basename(path) + ".tar")
        logger.info("Creating tarball from directory %s --> %s", path, outfile)
        if not dry:
            with switchdir(path):
                run_command(
                    [
                        "tar",
                        "-cvf",
                        outfile,
                        ".",
                        "--exclude",
                        "*/lib/python*",
                        "--exclude",
                        "*/include/python*",
                        "--exclude",
                        "*/bin/python*",
                        "--exclude",
                        "*.egg-info*",
                        "--exclude",
                        "*.pyc",
                        "--exclude",
                        "*/.git",
                        "--exclude",
                        "*/dist/*",
                        "--exclude",
                        "*/.fcache/*",
                        "--exclude",
                        "*/examples/*",
                    ]
                )
    else:
        logger.info("Package %s: Using top level git to create a tarball", path)
        # Get the top-level git dir
        with switchdir(path):
            toplevel_git_dir = run_command(["git", "rev-parse", "--show-toplevel"])[
                0
            ].strip()
        # Fix the output name of the tarball
        outfile = osp.join(outdir, osp.basename(toplevel_git_dir) + ".tar")
        with switchdir(toplevel_git_dir):
            if allow_uncommitted:
                logger.info(
                    "Creating tarball for %s including uncommitted changes",
                    toplevel_git_dir,
                )
                # Create the tarball with uncommitted changes in it
                if not dry:
                    run_command(
                        "git ls-files -z | xargs -0 tar -cvf {0}".format(outfile),
                        shell=True,
                    )
            else:
                # Check if there are uncommitted changes
                try:
                    run_command(["git", "diff-index", "--quiet", "HEAD", "--"])
                except subprocess.CalledProcessError:
                    logger.error(
                        "Uncommitted changes detected; it is unlikely you want a tarball "
                        "with some changes not committed."
                    )
                    raise
                # Create the actual tarball of the latest commit
                if not dry:
                    run_command(["git", "archive", "-o", outfile, "HEAD"])
    logger.info("Created tarball {0}".format(outfile))
    return outfile


def extract_tarball(tarball, outdir=".", verbose=False):
    """
    Extracts a tarball to outdir
    """
    tarball = osp.abspath(tarball)
    outdir = osp.abspath(outdir)
    create_directory(outdir)
    logger.warning("Extracting {0} ==> {1}".format(tarball, outdir))
    cmd = ["tar", "-x{}f".format("v" if verbose else ""), tarball, "-C", outdir]
    run_command(cmd)


def extract_tarball_cmssw(tarball, outdir="."):
    """
    Extracts a tarball to outdir, and returns the extracted CMSSW dir
    """
    extract_tarball(tarball, outdir)
    # return the CMSSW directory
    if qondor.DRYMODE:
        return "CMSSW_dry"
    # Get the extracted directory from the tarball:
    import tarfile

    for name in tarfile.open(tarball).getnames():
        if name.startswith("CMSSW"):
            break
    else:
        raise RuntimeError(
            'Could not find any directory in {} that starts with "CMSSW"'.format(
                tarball
            )
        )
    return osp.join(outdir, name)


def check_is_cmssw_path(path):
    """
    Checks whether the passed path contains a CMSSW distribution.
    """
    abs_path = osp.abspath(path)
    if not osp.basename(path).startswith("CMSSW"):
        raise ValueError(
            'Expected {0} to start with "CMSSW" (path: {1})'.format(
                osp.basename(path), abs_path
            )
        )
    if not osp.isdir(path):
        raise OSError("{0} is not a directory (path: {1})".format(path, abs_path))
    if not osp.isdir(osp.join(path, "src")):
        raise OSError(
            "{0} is not a directory (path: {1})".format(osp.join(path, "src"), abs_path)
        )


def get_clean_env():
    env = os.environ.copy()
    for var in [
        "ROOTSYS",
        "PATH",
        "LD_LIBRARY_PATH",
        "DYLD_LIBRARY_PATH",
        "SHLIB_PATH",
        "LIBPATH",
        "PYTHONPATH",
        "MANPATH",
        "CMAKE_PREFIX_PATH",
        "JUPYTER_PATH",
        # Added due to ROOT-env.sh
        "CPLUS_INCLUDE_PATH",
        "CXX",
        "ZLIB_HOME",
        "CURL_HOME",
        "DAVIX_HOME",
        "GSL_HOME",
        "SETUPTOOLS_HOME",
        "FONTCONFIG_HOME",
        "CAIRO_HOME",
        "SQLITE_HOME",
        "PIXMAN_HOME",
        "FREETYPE_HOME",
        "TBB_HOME",
        "FC",
        "PKG_CONFIG_HOME",
        "VC_HOME",
        "PNG_HOME",
        "FFTW_HOME",
        "BOOST_HOME",
        "VDT_HOME",
        "ROOT_HOME",
        "ZEROMQ_HOME",
        "LIBXML2_HOME",
        "PKG_CONFIG_PATH",
        "EXPAT_HOME",
        "COMPILER_PATH",
        "BLAS_HOME",
        "R_HOME",
        "XROOTD_HOME",
        "MYSQL_HOME",
        "GFAL_HOME",
        "CC",
        "C_INCLUDE_PATH",
        "PYTHON_HOME",
        "PYTHONHOME",
        "ORACLE_HOME",
        "GPERF_HOME",
        "SRM_IFCE_HOME",
        "NUMPY_HOME",
        "DCAP_HOME",
    ]:
        if var in env:
            del env[var]
    return env


def convert_to_utc(local_time):
    """
    Converts a local time to UTC.
    Implemented for only a few basic timezones.
    Needs to be extended with pytz if this function
    get seriously used.
    """
    # See: https://stackoverflow.com/a/10854983/9209944
    delta = datetime.timedelta(
        seconds=time.timezone if (time.localtime().tm_isdst == 0) else time.altzone
    )
    new_time = local_time + delta
    logger.debug(
        "Offsetting %s by %s hours to get to UTC %s", local_time, delta, new_time
    )
    return new_time


def get_now_utc():
    return convert_to_utc(datetime.datetime.now())


def sleep_until(runtime_utc, allowed_lateness=300, is_not_utc=False):
    if is_not_utc:
        runtime_utc = convert_to_utc(runtime_utc)
    now_utc = get_now_utc()
    logger.info("Current time (UTC):       %s", now_utc)
    logger.info("Scheduled run time (UTC): %s", runtime_utc)

    delta = runtime_utc - now_utc
    delta_seconds = abs(delta.total_seconds())

    if delta < datetime.timedelta(seconds=0):
        # The job is too late; runtime_utc has already passed
        if delta_seconds < allowed_lateness:
            logger.info(
                "Job is late by %s seconds, which is within the allowed window"
                " of max %s seconds late",
                delta_seconds,
                allowed_lateness,
            )
            return 0
        else:
            logger.error(
                "Job is late by %s seconds, which is OUTSIDE the allowed window"
                " of max %s seconds late - throwing exception",
                delta_seconds,
                allowed_lateness,
            )
            raise RuntimeError
    else:
        logger.info("Job is early by %s seconds, sleeping", delta_seconds)
        time.sleep(delta_seconds)
        return 0


def check_proxy():
    """
    Asserts that the user has a grid proxy that is valid for at least 168 more hours (1 week)
    """
    # cmd = 'voms-proxy-info -exists -valid 168:00' # Check if there is an existing proxy for a full week
    if qondor.DRYMODE:
        return
    try:
        subprocess.check_output(["grid-proxy-info", "-exists", "-valid", "168:00"])
        logger.info("Found a valid proxy")
    except subprocess.CalledProcessError:
        logger.error(
            "Grid proxy is not valid for at least 1 week. Renew it using:\n"
            "voms-proxy-init -voms cms -valid 192:00"
        )
        raise
    except OSError:
        logger.error(
            "The grid-proxy-info is not found or behaves weird; "
            "Continuing but note that your grid proxy may not be valid"
        )
        pass


def dist_is_editable(dist):
    """
    Is distribution an editable install?
    see: https://stackoverflow.com/a/42583363/9209944
    """
    # If a string is passed, convert it to a module object
    if is_string(dist):
        import pkg_resources

        try:
            dist = pkg_resources.get_distribution(dist)
        except pkg_resources.DistributionNotFound:
            logger.info(
                "Package %s is not installed; assuming non-editable install", dist
            )
            return False
    # Check if the module's .egg-link is somewhere on the python path
    for path_item in sys.path:
        egg_link = osp.join(path_item, dist.project_name + ".egg-link")
        if osp.isfile(egg_link):
            return True
    return False


def _iter_chunkify_nrange(list_length, n_chunks):
    """
    Makes n_chunks chunks out of a range(list_length) list.
    Returns empty lists if n_chunks > list_length.
    """
    n_per_chunk_f = float(list_length) / n_chunks
    boundaries = [(i * n_per_chunk_f, (i + 1) * n_per_chunk_f) for i in range(n_chunks)]
    for left, right in boundaries:
        indices_in_chunk = []
        for i in range(list_length):
            if i >= left and i < right:
                indices_in_chunk.append(i)
        yield indices_in_chunk


def iter_chunkify(mylist, n_chunks):
    """
    Makes n_chunks chunks out of mylist.
    """
    for indices in _iter_chunkify_nrange(len(mylist), n_chunks):
        yield [mylist[i] for i in indices]


def chunkify(mylist, n_chunks=None, chunksize=None):
    """
    Must specify either n_chunks or chunksize. Returns chunks of mylist as a list.
    """
    if (n_chunks is None and chunksize is None) or (
        n_chunks is not None and chunksize is not None
    ):
        raise Exception("Must specify either n_chunks or chunksize")
    if chunksize is not None:
        n_chunks = int(math.ceil(len(mylist) / float(chunksize)))
    return list(iter_chunkify(mylist, n_chunks))


def get_ith_chunk(mylist, n_chunks, i_chunk):
    return chunkify(mylist, n_chunks)[i_chunk]
